﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Collections.ObjectModel;
using System.Reflection;
using System.CodeDom.Compiler;
using Microsoft.CSharp;
using System.Collections.Specialized;
using System.ComponentModel;

namespace ScorchedEarth
{
    public class Expression : INotifyPropertyChanged
    {
        private ObservableCollection<Parameter> parameters;
        public ObservableCollection<Parameter> Parameters
        {
            get
            {
                return parameters;
            }
            set
            {
                if (parameters != null && parameters != value)
                {
                    parameters.CollectionChanged -= new NotifyCollectionChangedEventHandler(Parameters_CollectionChanged);
                    foreach (var param in parameters)
                    {
                            param.PropertyChanged -= new System.ComponentModel.PropertyChangedEventHandler(param_PropertyChanged);
                    }

                    parameters = value;
                    Parameters.CollectionChanged += new NotifyCollectionChangedEventHandler(Parameters_CollectionChanged);
                    foreach (var param in parameters)
                    {
                        param.PropertyChanged += new System.ComponentModel.PropertyChangedEventHandler(param_PropertyChanged);
                    }
                    OnPropertyChanged("Parameters");
                }
            }
        }
        private string name = "Unknown";
        public string Name
        {
            get
            {
                return name;
            }
            set
            {
                if (value != null && name != value)
                {
                    name = value;
                    OnPropertyChanged("Name");
                }
            }
        }
        private string statement = String.Empty;
        public string Statement
        {
            get
            {
                return statement;
            }
            set
            {
                statement = value;
                BuildFunc();
                OnPropertyChanged("Statement");
            }
        }

        public string CompilerMessage { get; set; }
        public Type ClassType { get; private set; }
        public Func<double, double> Function { get; private set; }

        public Expression()
        {
            parameters = new ObservableCollection<Parameter>();
            parameters.CollectionChanged += new NotifyCollectionChangedEventHandler(Parameters_CollectionChanged);
        }

        #region INotifyPropertyChanged Members
        public event PropertyChangedEventHandler PropertyChanged;
        #endregion

        protected void OnPropertyChanged(string name)
        {
            PropertyChangedEventHandler handler = PropertyChanged;
            if (handler != null)
            {
                handler(this, new PropertyChangedEventArgs(name));
            }
        }

        void Parameters_CollectionChanged(object sender, System.Collections.Specialized.NotifyCollectionChangedEventArgs e)
        {
            if (e.Action == NotifyCollectionChangedAction.Add || e.Action == NotifyCollectionChangedAction.Replace)
            {
                foreach (var item in e.NewItems)
                {
                    Parameter param = item as Parameter;
                    if (param != null)
                        param.PropertyChanged += new System.ComponentModel.PropertyChangedEventHandler(param_PropertyChanged);
                }
            }
            else if (e.Action == NotifyCollectionChangedAction.Remove || e.Action == NotifyCollectionChangedAction.Replace)
            {
                foreach (var item in e.OldItems)
                {
                    Parameter param = item as Parameter;
                    if (param != null)
                        param.PropertyChanged -= new System.ComponentModel.PropertyChangedEventHandler(param_PropertyChanged);
                }
            }
            BuildFunc();
            OnPropertyChanged("Parameters");
        }

        void param_PropertyChanged(object sender, System.ComponentModel.PropertyChangedEventArgs e)
        {
            Parameter param = sender as Parameter;
                if (param != null)
            {
                if (e.PropertyName == "Name")
                    BuildFunc();
                else
                    // Change the value;
                    SetProperty(param.Name, param.Value);
                OnPropertyChanged("Parameters");
            }
        }

        #region Compiling, etc.
        private void SetProperty(string name, double value)
        {
            if (ClassType != null)
            {
                PropertyInfo staticProperty = ClassType.GetProperty(name);
                if (staticProperty != null)
                    staticProperty.SetValue(null, value, null);
            }

        }

        private void BuildFunc()
        {
            string sourceCode = AddExpression();
            try
            {
                CompileExpression(this.Name, sourceCode);
                this.CompilerMessage = "Compiled Successfully.";
                foreach (var param in Parameters)
                    SetProperty(param.Name, param.Value);
            }
            catch (ApplicationException exception)
            {
                this.CompilerMessage = "Failed! " + exception.Message;
            }
        }

        private string AddExpression()
        {
            //string declaration = "double func" + this.Name + "(double x) {";
            StringBuilder classDefinition = new StringBuilder();
            classDefinition.AppendLine("using System;");
            classDefinition.Append("public static class ");
            classDefinition.Append(this.Name);
            classDefinition.AppendLine(" {");
            foreach (Parameter param in this.Parameters)
            {
                classDefinition.Append("   public static double ");
                classDefinition.AppendLine(param.Name);
                classDefinition.AppendLine("      { get; set; }");
            }
            classDefinition.AppendLine("   public static double Height(double x) {");
            classDefinition.AppendLine("      double height = 0;");
            //classDefinition.AppendLine(");");
            classDefinition.AppendLine(this.Statement);
            classDefinition.AppendLine("      return height;");
            classDefinition.AppendLine("}   }");
            return classDefinition.ToString();
        }

        private void CompileExpression(string Name, string stringExpression)
        {
            Assembly assembly = CompileSource(stringExpression);
            //
            // Try to Invoke a method
            //
            this.ClassType = assembly.GetType(Name);
            MethodInfo heightMethod = this.ClassType.GetMethod("Height");
            this.Function = (Func<double, double>)Delegate.CreateDelegate(typeof(Func<double, double>), heightMethod);
        }

        private static Assembly CompileSource(string sourceCode)
        {
            CodeDomProvider cpd = new CSharpCodeProvider();
            CompilerParameters cp = new CompilerParameters();
            cp.ReferencedAssemblies.Add("System.dll");
            //cp.ReferencedAssemblies.Add("ClassLibrary1.dll");
            cp.GenerateExecutable = false;

            // Invoke compilation.
            CompilerResults cr = cpd.CompileAssemblyFromSource(cp, sourceCode);
            if (cr.Errors.HasErrors)
            {
                StringBuilder errors = new StringBuilder("Errors in compiling.");
                errors.Append(Environment.NewLine);
                foreach (CompilerError ce in cr.Errors)
                {
                    errors.AppendLine(ce.ToString());

                }

                throw new ApplicationException(errors.ToString());
            }
            return cr.CompiledAssembly;
        }
        #endregion
    }
}
